from soaplib.service import rpc
from soaplib.service import DefinitionBase
from soaplib.serializers.primitive import *
from lxml import etree
import cElementTree as et

from soaplib.serializers.binary import Attachment
from soaplib.wsgi import Application
from soaplib.serializers.clazz import *
import shutil

import sys
import os
import time
import hashlib
import base64
import MySQLdb

### Mark: Need a hostname field
class Module(ClassSerializer):
    """The Class which holds Module object"""
    __namespace__ = "user2"
    modulename = String
    arguments = String
    executableid = Integer
    relativeinputfiles = Array(Integer)
    relativeoutputfiles = Array(Integer)
 
class File(ClassSerializer):
    """The Class which holds File Object"""
    __namespace__ = "user3"
    filename = String
    metadata = String
    mimetype = String
    filesize = long

class Workflow(ClassSerializer):
    """The Class which holds the Workflow Object"""
    __namespace__ = "user1"
    filenames = Array(String)
    filedata = Array(String)
    label = String
    modules = Array(Module)
    metafiles = Array(File)

class WorkflowExecution(DefinitionBase):
    """The Class which does all the Workflow Execution and Generation Work"""

    @rpc(String, String, _returns=String)
    def getSecurityToken(self, username, password):
        """ Composes the Security Token 
	The security token is passed for each call
	this functin will return a security token for username/password

	@param username String
	@param password String
	@return String the securitytoken
	"""
	return hashlib.md5("%s:%s"%(username,password)).hexdigest()
	
    @rpc(Integer, Integer, String, String, _returns=Integer)
    def addDispatch(self, groupid, access, callback, securitytoken):
        """ Add dispatch id
	Prepares a new dispatch with dispatchid

	@param groupid Integer
	@param access Integer
	@param callback String
	@param securitytoken String
	@return Integer|None the dispatch id  
	"""
	try:
            dispatchid = 0
	    conns = MySQLdb.connect("localhost", "harsha", "swamp", "swampbackend")
	    cursors = conns.cursor()
	    cursors.execute("SELECT * FROM `session` WHERE token ='%s'" %(securitytoken))
	    results = cursors.fetchone()
	    if results!= None:
		userid = results[1]
		date = int(time.time())
		cursors.execute(""" INSERT INTO `dispatch` (groupid, userid, access, callback, date)\
                                    VALUES ('%d', '%d', '%d', '%s', '%d')"""\
                                    %(groupid,  userid, access, callback, date)) 
		dispatchid = conns.insert_id()
	    	os.chdir("/home/condor/MySoft/sample")
		s = str(dispatchid)
	 	os.mkdir(s)
		conns.commit()

	except MySQLdb.Error, e:
	    conns.rollback()
	    print "Error %d: %s" %(e.args[0], e.args[1])

	finally:    # executes this block even if error occurs or not
	    cursors.close()
	    conns.close()
            return dispatchid

### MARK: The hostname item in the module Table must be updated	
    def addModules(self, workflowid, module, files, cursors, conns):
	""" Adds modules with the workflowid.
	
	Adds the files and also module io relationships.
	@param workflowid Integer
	@param modules Array(Module)
	@param files Array(File)
	@param securitytoken String
	@return Integer[] the module ids or empty on error
	"""
	fileid = []
	moduleio = []
	insertid = []
	try:

	# Inserting the File Object details
### MARK: Use executemany
	    T = []
            for index, file in enumerate(files):
	        t = (file.filename, file.metadata, file.mimetype, file.filesize)
		T.append(t)
                # insert the file object details into DB
            cursors.executemany("INSERT INTO `file` (filename, metadata, mimetype, filesize)\
                                   VALUES ('%s', '%s', '%s', '%s')", T)
            p = conns.insert_id()
	    for index in range(T):
	        fileid.append(p-index)

	# Creating the data structure for storing the moduleio relationships
	    for i in range(len(fileid)):
                moduleio[fileid[i]] = {}
                moduleio[fileid[i]][src] = []
                moduleio[fileid[i]][dest] = []

### MARK: Use executemany	
	# Inserting Module Object information
	    T = []
	    for index, modules in enumerate(module):
		t = (workflowid, modules.modulename, modules.arguments, modules.executableid)
		T.append(t)

	    cursors.executemany(" INSERT INTO `module` (workflowid, modulename, arguments, executableid)\
                                    		          VALUES ('%s', '%s', '%s', '%s')", T) # insert module details
### NEED to be test
            p = conn.insert_id()
	    for index in range(T):
		insertid[modules].append(p-index)
					
#building the dependecies
		for i, v in enumerate(modules.relativeinputfiles):	
		    moduleio[fileid[v]][dest].append(modules)

		for i, v in enumerate(modules.relativeoutputfiles):
		    moduleio[fileid[v]][src].append(modules)
				
### MARK: Use executemany
	    T= []	
	    for index, value in enumerate(fileid):
                for i, v in enumerate(moduleio[fileid[index]][src]):
		    for x, y in enumerate(moduleio[fileid[index]][dest]):
		        t = (value, insertid[moduleio[fileid[index]][src][i]], insertid[moduleio[fileid[index]][dest][x]])
			T.append(t)
					
	    cursors.executemany(" INSERT INTO `moduleio` (fileid, srcmoduleid, destmoduleid)\
                                            VALUES ('%s', '%s', '%s') ", T)
		
            conns.commit()    # saves teh changes made to the database

        except MySQLdb.Error, e:
	    conns.rollback()
	    print "Error %d: %s" %(e.args[0], e.args[1])
            insertid = []

	finally:
            return insertid

    @rpc(Integer, Array(Workflow), String, _returns=Array(Integer))
    def addWorkflows(self, dispatchid, workflows, securitytoken):
        """ Add Workflows to dispatchid.
	
	Prepares new Workflows with specified files
	@param dispatchid Integer
	@param workflows Array(Workflow)
	@param securitytoken String
	@return Integer[] the worklfow ids or empty on error
	"""		
	wids = []
	try:
	    conns = MySQLdb.connect("localhost", "harsha", "swamp", "swampbackend")
	    cursors = conns.cursor()
            
            # To check if the dispatch to which we are going to add the workflow exists or not
	    cursors.execute("SELECT * FROM `dispatch` a, `session` b\
                             WHERE a.userid = b.userid and b.token = '%s' and a.dispatchid = '%d'"\
                             %(securitytoken, dispatchid))
	    result1 = cursors.fetchone() 
	    
            if result1 != None:

		T = []	
	        for index, workflow in enumerate(workflows):
		    t = (dispatchid, workflow.label, int(time.time()))
		    T.append(t)
		

### MARK: Use executemany	
                    # creating the workflow for the dispatch of the user
		cursors.executemany(" INSERT INTO `workflow`\
                                      (dispatchid, label, submitdate, execdate, enddate, endstate)\
                                      VALUES ('%s', '%s', '%s', 0, 0, 0) ", T)

### MARK: How to get insertids ??????
		    # get back the recently inserted id 
                p = conn.insert_id()
		for index in range(T):
		    wids.append(p-index) 

	 	os.chdir("/home/condor/MySoft/sample")
		os.chdir(str(dispatchid))
		i = str(conn.insert_id())
		os.mkdir(i)
		os.chdir(i)
		for fileindex, filedata in enumerate(workflow.filedata):
		    fname = str(workflow.filenames[fileindex])
		    base64.decode(filedata, open(fname, "w"))
		    
		if workflow.modules !=  None:
		        self.addModules(wids[index], workflow.modules, workflow.metafiles, cursors, conns)
                
		conns.commit()

        except MySQLdb.Error, e:
	    conns.rollback()
	    print "Error %d: %s" %(e.args[0], e.args[1])
            wids = []

	finally:
	    cursors.close()
	    conns.close()	
            return wids

    @rpc(Integer, String, _return=Boolean)
    def removeDispatch(self, dispatchid, securitytoken):
        """ Remove dispatches with dispatchids.

	Removes all files, subworkflows, workflows, and dispatch itself with dispatchids.
	@param dispatchid Integer
	@param securitytoken String
	@return Boolean true on success, false on error
	"""
	try:
	    wid = []
	    ret = False
	    conns = MySQLdb.connect("localhost", "harsha", "swamp", "swampbackend")
	    cursors = conns.cursor()
	    cursors.execute(" SELECT * FROM `session` WHERE token = '%s'"%(securitytoken))
	    results = cursors.fetchone()
	    if results != None:
	        uid = results[1]
		cursors.execute("SELECT * FROM `dispatch` WHERE \
                                 dispatchid = %d and userid = %d"\
                                 %(dispatchid, uid))
		results = cursors.fetchone()
		if results != None:
		    cursors.execute("SELECT * FROM `workflow` WHERE dispatchid = %d" %(dispatchid)) 
		    workflowresult = cursors.fetchall()
		    if workflowresult != None:
		        for row6 in workflowresult:
		            wid.append(row6[0])	
		        self.removeWorkflows(wid, securitytoken)
		cursors.execute(" DELETE FROM `dispatch` WHERE dispatchid = %d and userid = %d" %(dispatchid, uid))
		conns.commit()
	# removing the directory of the dispatch
		os.chdir("/home/condor/MySoft/sample/")
		cmd = str(dispatchid)
		shutil.rmtree(cmd)
                ret = True

        except MySQLdb.Error, e:
	    conns.rollback()
            print "Error %d: %s" %(e.args[0], e.args[1])
	    ret = False

	finally:
            cursors.close()
            conns.close()
            return ret

    @rpc(Array(Integer), String, _return=Boolean)
    def removeWorkflows(self, workflowids, securitytoken):
        """ Remove Workflows of dispatchid.

	Removes all files and Workflow itself from worklfowid of dispatchid.
	@param dispatchid Integer
	@param workflowids Array(Integer)
	@param securitytoken String
	@return Boolean true on success, false on error
	"""
        try:
	    conns = MySQLdb.connect("localhost", "harsha", "swamp", "swampbackend")
	    cursors = conns.cursor()
	    cursors.execute("SELECT * FROM `session` WHERE token = '%s'" %(securitytoken))
	    results = cursors.fetchone()
	    if results != None:
	        userid = results[1]
 
### MARK use Patrick's SQL in the Email

            # removes workflows
                cursors.execute(" DELETE workflow.* FROM `workflow`, `dispatch`\
                                 WHERE workflow.workflowid IN %s AND\
		     	         workflow.dispatchid = dispatch.dispatchid AND\
                                 dispatch.userid = %d" %(workflowids, userid))

	    # To get the modules associated for that workflowid
	        moduleids = []
	        cursors.execute(" SELECT * FROM `module` WHERE workflowid IN %s", workflowids)
	        results = cursors.fetchone()
	        if results != None:
		    for row in results:
	  	        moduleids.append(row[0])
		
            # removes the module information of the workflow
                cursors.execute( "DELETE FROM `module`\
                                  WHERE workflowid IN %s", workflowids)
	        files = []
	    # To get the files associated for that modules
	        cursors.execute( " SELECT * FROM `moduleio`\
		    	           WHERE srcmoduleid  OR destmoduleid IN %s", moduleids)
	        results = cursors.fetchall()
	        if results != None:
	            for row in results:
		        files.append(row[1])

### MARK need to remove the moduleio
	    #remove the moduleio relation from DB
	        cursors.execute(" DELETE FROM `moduleio`\
			          WHERE srcmoduleid OR destmoduleid IN %s", moduleids)		

	        cursors.execute(" DELETE FROM `file`\
			          WHERE fileid IN %s", files)


	#Delete the folders too
	        os.chdir("/home/condor/MySoft/ws/")
	        g = str(value.dispatchid)

	        oschdir(g)
	        h = str(value)
	        shutil.rmtree(h)
		
	        conns.commit()
	        ret = True
	    else:
	        ret = False	

        except MySQLdb.Error, e:
	    conns.rollback()
	    print "Error %d: %s" %(e.args[0], e.args[1])
	    ret = False

	finally:
	    cursors.close()
	    conns.close()
            return ret

### MARK: A lot of changes to do
    def generation(self, dispatchid):
	""" Generates the submit file for workflows of dispatch.

	This function will create the submit file for each of the workflow.
	@param dispatchid Integer
	@return Boolean true, false on error
	"""
	try:
	    conns = MySQLdb.connect("localhost", "harsha", "swamp", "swampbackend")
	    cursors = conns.cursor()
	    cursors.execute("SELECT workflow.workflowid as workflowid, module.moduleid as moduleid,\
                             module.modulename as modulename, module.arguments as arguments,\
                             executable.filename as efilename, module.hostname as hostname,\
                             module.fileid as fileid,\
                             moduleio.srcmoduleid as srcmoduleid, moduleio.destmoduleid as destmoduleid,\
                             file.filename as ffilename\
                             FROM `workflow` LEFT JOIN (module)\
                             ON workflow.workflowid = module.workflowid\
                             LEFT JOIN (moduleio)\
                             ON module2.moduleid = moduleio.srcmoduleid\
                             OR module.moduleid = moduleio.destmoduleid\
                             LEFT JOIN (file)\
                             ON moduleio.fileid = file.fileid\
                             WHERE workflow.dispatch = '%d'\
                             GROUP BY workflow.workflowid, module.moduleid, moduleio.fileid\
                             ORDER BY NULL " %(dispatchid))
	    rows = cursors.fetchall()        
	    prevWorkflow = 0
	    prevModule = 0
	    lastRow = ()
	    if rows != None:
	        for row in rows:
		#Handle Workflow	
	            if (row[0] != prevWorkflow):
		        if (prevWorkflow != 0):
#creating a dag file for the module
		            for parent, children in dependency:			
			        filedata += "PARENT %s CHILD"%(parent)
			        for child in children:
			            filedata += child + ""
			#write filedata to .dag file
			    dagname = workflowid + ".dag"
			    dHandler = open(dagname, 'w')
			    dHandler.write(filedata)
			    
		        prevWorkflow = row[0]
		        dependency = {}

		       # Handle module
                    if (row[1] != prevModule):
		        if (prevModule != 0):
			# creating "submit" file and "stork" file					
		            fname = ''
			    fname = row[1] + '.submit'	
			    fileHandler = open(fname,'w')
			    fileHandler.write('####################\n # submit file for \t%d' %(row[0]))
			    fileHandler.write('\n####################\n\n Universe\t=\tvanilla \nExecutable\t=\t%s' %(row[2]))
			    fileHandler.write('\nshould_transfer_files\t=\tYES')
			    fileHandler.write('\nwhen_to_transfer_output\t=\tON_EXIT')
			    fileHandler.write('\nrequirements\t=\t%s' %(row[5]))
			    fileHandler.write('\n\nQueue')

			    sname = ''
			    sname = row[2] + '.stork'
			    sHandler = open(sname, 'w')
			    sHandler.write('[ \n dap_type\t= \t"transfer"; \n src_url\t=\t%s' %(row[7]))
			    sHandler.write('\ndest_url\t=\%s' %(row[5]))

			prevModule = row[1]
			dependency[row[1]] = {}
			

		    inputfiles = []
		    outputfiles = []
					
	# Handle files
	# if this is a output file
		    if (row[8]) == row[1]:
		        outputfiles.append(row[4])

		    if (row[9] != None):
		        dependency[row[1]][row[9]] = 1

	# if this is an input file
		    if (row[9] == row[1]):
		        inputfiles.append(row[4])

		    lastRow = row

	            return True
	        else:
		    return False

	except MySQLdb.Error, e:
	    print "Error %d: %s" %(e.args[0], e.args[1])
	    return False

	finally:
	    cursors.close()
	    conns.close()

    @rpc(Integer, String, String, _return=Boolean)
    def dispatch(self, dispatchid, callbackuri, securitytoken):
    	""" Dispatch the prepared dispatch id.
	
	This function will dispatch the prepared dispatchid to condor system.
	A supplied callback uri is supplied for post-dispatch communication of dispatch status, statistics, results
	@param dispatchid Integer
	@param callbackuri String
	@param securitytoken String
	@return Boolean true, false in error
	"""
	try:
	    conns = MySQLdb.connect("localhost", "harsha", "swamp", "swampbackend")
	    cursors = conns.cursor()
	    cursors.execute("SELECT * FROM `session` WHERE token = '%s'" %(securitytoken))
	    results = cursors.fetchone()
	    if results != None:
		userid = results[1]
                # call the generation function to create submit files
	        if self.generation(dispatchid): 
### MARK: submit dag file for all the workflows
		    f = "condor_submit_dag " + str(dispatchid) + ".dag"
		    os.system(f) # sample command
	        ret = True
	    else:
		ret = False

	except MySQLdb.Error, e:
	    print "Error %d: %s" %(e.args[0], e.args[1])
	    ret = False

	finally:
	    cursors.close()
	    conns.close()
            return ret

if __name__=='__main__':
    try:
        from wsgiref.simple_server import make_server
	server = make_server('localhost', 7789, Application([WorkflowExecution], 'tns'))
	server.serve_forever()
    except ImportError:
	print "Error: server code requires Python >= 2.5"
